@@ -15,6 +15,7 @@
 #include "pos-emoji-picker.h"
 #include "pos-input-method.h"
 #include "pos-completer.h"
+#include "pos-completer-manager.h"
 #include "pos-completion-bar.h"
 #include "pos-input-surface.h"
 #include "pos-main.h"
@@ -38,11 +39,13 @@ enum {
   PROP_0,
   PROP_INPUT_METHOD,
   PROP_COMPLETER,
+  PROP_COMPLETER_MANAGER,
   PROP_SCREEN_KEYBOARD_ENABLED,
   PROP_KEYBOARD_DRIVER,
   PROP_SURFACE_VISIBLE,
   PROP_COMPLETER_ACTIVE,
   PROP_COMPLETION_ENABLED,
+  PROP_OSK_FEATURES,
   PROP_LAST_PROP
 };
 static GParamSpec *props[PROP_LAST_PROP];
@@ -86,6 +89,7 @@ struct _PosInputSurface {
   GtkWidget               *emoji_picker;
   GtkWidget               *last_layout;
   PosShortcutsBar         *shortcuts_bar;
+  PhoshOskFeatures         osk_features;
 
   /* The debug surface */
   PosDebugWidget          *debug_widget;
@@ -104,6 +108,7 @@ struct _PosInputSurface {
   /* word completion */
   GtkWidget               *word_completion_btn;
   PosCompleter            *completer;
+  PosCompleterManager     *completer_manager;
   GtkWidget               *completion_bar;
   gboolean                 completion_enabled;
   PhoshOskCompletionModeFlags completion_mode;
@@ -240,8 +245,10 @@ on_completer_preedit_changed (PosInputSurface *self)
 static void
 on_completer_completions_changed (PosInputSurface *self)
 {
+  g_auto (GStrv) completions = pos_completer_get_completions (self->completer);
+
   pos_completion_bar_set_completions (POS_COMPLETION_BAR (self->completion_bar),
-                                      pos_completer_get_completions (self->completer));
+                                      completions);
 }
 
 
@@ -537,7 +544,7 @@ select_layout_change_state (GSimpleAction *action,
 
 
 static void
-switch_language (PosInputSurface *self, const char *locale)
+switch_language (PosInputSurface *self, const char *locale, const char *region)
 {
   gboolean success;
   g_autoptr (GError) err = NULL;
@@ -545,12 +552,16 @@ switch_language (PosInputSurface *self, const char *locale)
   if (self->completer == NULL)
     return;
 
-  g_debug ("Switching language, locale: '%s'", locale);
-  success = pos_completer_set_language (self->completer, locale, &err);
+  g_debug ("Switching language, locale: '%s-%s'", locale, region);
+  success = pos_completer_set_language (self->completer, locale, region, &err);
   if (success == FALSE) {
-    g_warning ("Failed to set language: %s, switching to '%s' instead",
-               err->message, POS_COMPLETER_DEFAULT_LANG);
-    pos_completer_set_language (self->completer, POS_COMPLETER_DEFAULT_LANG, NULL);
+    g_warning ("Failed to set language: %s-%s: %s, switching to '%s-%s' instead",
+               locale, region, err->message, POS_COMPLETER_DEFAULT_LANG,
+               POS_COMPLETER_DEFAULT_REGION);
+    pos_completer_set_language (self->completer,
+                                POS_COMPLETER_DEFAULT_LANG,
+                                POS_COMPLETER_DEFAULT_REGION,
+                                NULL);
   }
 
   pos_completion_bar_set_completions (POS_COMPLETION_BAR (self->completion_bar), NULL);
@@ -577,7 +588,8 @@ on_visible_child_changed (PosInputSurface *self)
   /* Remember last layout */
   self->last_layout = GTK_WIDGET (osk);
 
-  switch_language (self, pos_osk_widget_get_locale (osk));
+  if (osk != POS_OSK_WIDGET (self->osk_terminal))
+    switch_language (self, pos_osk_widget_get_lang (osk), pos_osk_widget_get_region (osk));
 
   /* Recheck completion bar visibility */
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_COMPLETER_ACTIVE]);
@@ -689,7 +701,9 @@ pos_input_surface_set_completer (PosInputSurface *self, PosCompleter *completer)
                       "swapped-signal::update",
                       G_CALLBACK (on_completer_update), self,
                       NULL);
-    switch_language (self, pos_osk_widget_get_locale (POS_OSK_WIDGET (self->last_layout)));
+    switch_language (self,
+                     pos_osk_widget_get_lang (POS_OSK_WIDGET (self->last_layout)),
+                     pos_osk_widget_get_region (POS_OSK_WIDGET (self->last_layout)));
   } else {
     g_debug ("Removing completer");
   }
@@ -698,6 +712,23 @@ pos_input_surface_set_completer (PosInputSurface *self, PosCompleter *completer)
 }
 
 
+static void
+pos_input_surface_set_completer_manager (PosInputSurface     *self,
+                                         PosCompleterManager *completer_manager)
+{
+  PosCompleter *default_completer;
+
+  if (self->completer_manager == completer_manager)
+    return;
+
+  g_set_object (&self->completer_manager, completer_manager);
+
+  /* Update the current completer */
+  default_completer = pos_completer_manager_get_default_completer (self->completer_manager);
+  pos_input_surface_set_completer (self, default_completer);
+}
+
+
 static double
 reverse_ease_out_cubic (double t)
 {
@@ -721,6 +752,29 @@ pos_input_surface_set_completion_enabled (PosInputSurface *self, gboolean enable
 }
 
 
+static void
+update_osk_features (gpointer key, gpointer value, gpointer data)
+{
+  PosOskWidget *osk_widget = POS_OSK_WIDGET (value);
+  PosInputSurface *self = POS_INPUT_SURFACE (data);
+
+  pos_osk_widget_set_features (osk_widget, self->osk_features);
+}
+
+
+static void
+pos_input_surface_set_osk_features (PosInputSurface *self, PhoshOskFeatures osk_features)
+{
+  if (self->osk_features == osk_features)
+    return;
+
+  self->osk_features = osk_features;
+  g_hash_table_foreach (self->osks, update_osk_features, self);
+
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_OSK_FEATURES]);
+}
+
+
 static void
 pos_input_surface_set_property (GObject      *object,
                                 guint         property_id,
@@ -734,8 +788,8 @@ pos_input_surface_set_property (GObject      *object,
     self->input_method = g_value_dup_object (value);
     pos_debug_widget_set_input_method (self->debug_widget, self->input_method);
     break;
-  case PROP_COMPLETER:
-    pos_input_surface_set_completer (self, g_value_get_object (value));
+  case PROP_COMPLETER_MANAGER:
+    pos_input_surface_set_completer_manager (self, g_value_get_object (value));
     break;
   case PROP_SCREEN_KEYBOARD_ENABLED:
     pos_screen_keyboard_set_enabled (self, g_value_get_boolean (value));
@@ -749,6 +803,9 @@ pos_input_surface_set_property (GObject      *object,
   case PROP_COMPLETION_ENABLED:
     pos_input_surface_set_completion_enabled (self, g_value_get_boolean (value));
     break;
+  case PROP_OSK_FEATURES:
+    pos_input_surface_set_osk_features (self, g_value_get_flags (value));
+    break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     break;
@@ -768,6 +825,9 @@ pos_input_surface_get_property (GObject    *object,
   case PROP_COMPLETER:
     g_value_set_object (value, self->completer);
     break;
+  case PROP_COMPLETER_MANAGER:
+    g_value_set_object (value, self->completer_manager);
+    break;
   case PROP_SCREEN_KEYBOARD_ENABLED:
     g_value_set_boolean (value, pos_input_surface_get_screen_keyboard_enabled (self));
     break;
@@ -780,6 +840,9 @@ pos_input_surface_get_property (GObject    *object,
   case PROP_COMPLETION_ENABLED:
     g_value_set_boolean (value, self->completion_enabled);
     break;
+  case PROP_OSK_FEATURES:
+    g_value_set_flags (value, self->osk_features);
+    break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     break;
@@ -962,6 +1025,7 @@ pos_input_surface_finalize (GObject *object)
   g_clear_object (&self->xkbinfo);
   g_clear_object (&self->css_provider);
   g_clear_object (&self->completer);
+  g_clear_object (&self->completer_manager);
   g_clear_pointer (&self->theme_name, g_free);
   g_clear_pointer (&self->osks, g_hash_table_destroy);
 
@@ -1162,14 +1226,25 @@ pos_input_surface_class_init (PosInputSurfaceClass *klass)
                          POS_TYPE_INPUT_METHOD,
                          G_PARAM_READWRITE |
                          G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
+   /**
+    * PosInputSurface:completer:
+    *
+    * A completer implementing the #PosCompleter interface.
+    *
+    */
+   props[PROP_COMPLETER] =
+     g_param_spec_object ("completer", "", "",
+                          POS_TYPE_COMPLETER,
+                          G_PARAM_READABLE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
   /**
-   * PosInputSurface:completer:
+   * PosInputSurface:completer-manager:
    *
-   * A completer implementing the #PosCompleter interface.
+   * A completer manager to use. It allows to query which implementation of
+   * [iface@Completer] to use.
    */
-  props[PROP_COMPLETER] =
-    g_param_spec_object ("completer", "", "",
-                         POS_TYPE_COMPLETER,
+  props[PROP_COMPLETER_MANAGER] =
+    g_param_spec_object ("completer-manager", "", "",
+                         POS_TYPE_COMPLETER_MANAGER,
                          G_PARAM_READWRITE |
                          G_PARAM_CONSTRUCT_ONLY | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
   /**
@@ -1198,7 +1273,6 @@ pos_input_surface_class_init (PosInputSurfaceClass *klass)
   props[PROP_COMPLETER_ACTIVE] =
     g_param_spec_boolean ("completer-active", "", "", FALSE,
                           G_PARAM_READABLE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
-
   /**
    * PosInputSurface:completion-enabled
    *
@@ -1215,6 +1289,16 @@ pos_input_surface_class_init (PosInputSurfaceClass *klass)
                          /* TODO: should be an interface */
                          POS_TYPE_VK_DRIVER,
                          G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS);
+  /**
+   * PosInputSurface:osk-features
+   *
+   * Features to enable on all OSKs.
+   */
+  props[PROP_OSK_FEATURES] =
+    g_param_spec_flags ("osk-features", "", "",
+                        PHOSH_TYPE_OSK_FEATURES,
+                        PHOSH_OSK_FEATURE_DEFAULT,
+                        G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
   g_object_class_install_properties (object_class, PROP_LAST_PROP, props);
 }
@@ -1234,7 +1318,7 @@ insert_osk (PosInputSurface *self,
   if (osk_widget)
     return osk_widget;
 
-  osk_widget = pos_osk_widget_new ();
+  osk_widget = pos_osk_widget_new (self->osk_features);
   if (!pos_osk_widget_set_layout (POS_OSK_WIDGET (osk_widget),
                                   display_name, layout, variant, &err)) {
     g_warning ("Failed to load osk layout for %s: %s", name, err->message);
@@ -1411,7 +1495,8 @@ pos_input_surface_init (PosInputSurface *self)
   g_settings_bind (self->a11y_settings, "screen-keyboard-enabled",
                    self, "screen-keyboard-enabled", G_SETTINGS_BIND_GET);
 
-  self->osks = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, (GDestroyNotify)gtk_widget_destroy);
+  self->osks = g_hash_table_new_full (g_str_hash, g_str_equal, g_free,
+                                      (GDestroyNotify)gtk_widget_destroy);
   self->xkbinfo = gnome_xkb_info_new ();
   self->input_settings = g_settings_new ("org.gnome.desktop.input-sources");
   self->osk_settings = g_settings_new ("sm.puri.phosh.osk");
@@ -1419,6 +1504,7 @@ pos_input_surface_init (PosInputSurface *self)
                             G_CALLBACK (on_completion_mode_changed),
                             self);
   on_completion_mode_changed (self, NULL, self->osk_settings);
+  g_settings_bind (self->osk_settings, "osk-features", self, "osk-features", G_SETTINGS_BIND_GET);
 
   pos_osk_widget_set_layout (POS_OSK_WIDGET (self->osk_terminal),
                              _("Terminal"),
